# Tensorflow re-implementaiton of: https://github.com/theislab/dca
# License: https://github.com/theislab/dca/blob/master/LICENSE.txt
# Modification by Trung Ngo 2019
from __future__ import absolute_import, division, print_function

from typing import Iterable

import tensorflow as tf
from tensorflow.python.keras.layers import Dense, Layer

from odin.bay.helpers import Statistic
from odin.networks import (DenseDeterministic, DenseDistribution, Identity,
                           Parallel)
from sisua.models.base import SingleCellModel
from sisua.models.networks import DenseNetwork


class DeepCountAutoencoder(SingleCellModel):
  """ Deep Count Autoencoder

  """

  def __init__(self,
               units,
               dispersion='full',
               xdist='zinb',
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               biased_latent=False,
               nlayers=2,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    super(DeepCountAutoencoder, self).__init__(units=units,
                                               xdist=xdist,
                                               dispersion=dispersion,
                                               parameters=locals(),
                                               **kwargs)
    self.encoder = DenseNetwork(n_units=hdim,
                                nlayers=nlayers,
                                activation='relu',
                                batchnorm=batchnorm,
                                input_dropout=xdrop,
                                output_dropout=edrop,
                                seed=self.seed,
                                name='Encoder')
    if linear_decoder:
      self.decoder = Identity(name="Decoder")
    else:
      self.decoder = DenseNetwork(n_units=hdim,
                                  nlayers=nlayers,
                                  activation='relu',
                                  batchnorm=batchnorm,
                                  input_dropout=zdrop,
                                  output_dropout=ddrop,
                                  seed=self.seed,
                                  name='Decoder')
    self.latent_layer = DenseDeterministic(zdim,
                                           use_bias=bool(biased_latent),
                                           activation='linear',
                                           name='Latent')
    # initialize output_layers distribution
    # we must set the attribute directly so the Model will manage
    # the output layer, once all the output layers are initialized
    # the number of outputs will match `n_outputs`
    for idx, (units, posterior,
              activation) in enumerate(zip(self.units, self.xdist,
                                           self.xactiv)):
      name = 'output_layer%d' % idx
      post = DenseDistribution(
          units=units,
          posterior=posterior,
          activation=activation,
          posterior_kwargs=dict(dispersion=self.dispersion))
      setattr(self, name, post)

  def _call(self, x, lmean, lvar, t, y, masks, training, n_samples):
    e = self.encoder(x, training=training)
    qZ = self.latent_layer(e)
    # the first dimension always the MCMC sample dimension
    d = self.decoder(qZ.sample(1), training=training)

    pX = [
        getattr(self, 'output_layer%d' % i)(d, mode=Statistic.DIST)
        for i in range(self.n_outputs)
    ]

    # calculating the losses
    loss_x = self.xloss[0](t, pX[0])
    # don't forget to apply mask for semi-supervised loss
    loss_y = tf.convert_to_tensor(0, dtype=x.dtype)
    for i_true, m, i_pred, fn_loss in zip(y, masks, pX[1:], self.xloss[1:]):
      loss_y += fn_loss(i_true, i_pred) * m
    loss = tf.reduce_mean(loss_x + loss_y)

    if training:
      self.add_loss(loss)
    self.add_metric(loss_x, 'mean', "loss_x")
    if self.is_semi_supervised:
      self.add_metric(loss_y, 'mean', "loss_y")

    return pX, qZ



from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from numbers import Number
from typing import List, Union

import numpy as np
import tensorflow as tf
from six import add_metaclass
from tensorflow.python.keras.callbacks import Callback
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python.distributions import Distribution

from odin.bay.distributions import ZeroInflated
from odin.utils import catch_warnings_ignore
from sisua.analysis.imputation_benchmarks import (correlation_scores,
                                                  imputation_mean_score,
                                                  imputation_score,
                                                  imputation_std_score)
from sisua.analysis.latent_benchmarks import clustering_scores
from sisua.data import SingleCellOMIC
from sisua.models import SingleCellModel
from sisua.models.base import _to_sc_omics

__all__ = [
    'SingleCellMetric', 'NegativeLogLikelihood', 'ImputationError',
    'CorrelationScores', 'ClusteringScores'
]


def _preprocess_output_distribution(y_pred):
  """ In case of zero inflated distribution, extract the underlying count
  distribution """
  if isinstance(y_pred, tfd.Independent) and \
    isinstance(y_pred.distribution, ZeroInflated):
    y_pred = tfd.Independent(
        y_pred.distribution.count_distribution,
        reinterpreted_batch_ndims=y_pred.reinterpreted_batch_ndims)
  return y_pred


def _to_binary(protein):
  labels = protein.X
  if 'X_prob' in protein.obsm:
    labels = protein.obsm['X_prob']
  elif 'X_bin' in protein.obsm:
    labels = protein.obsm['X_bin']
  if labels.ndim == 2:
    labels = np.argmax(labels, axis=1)
  elif labels.ndim > 2:
    raise RuntimeError("protein labels has %d dimensions, no support" %
                       labels.ndim)
  return labels


_CORRUPTED_INPUTS = {}


# ===========================================================================
# Base class
# ===========================================================================
@add_metaclass(ABCMeta)
class SingleCellMetric(Callback):
  """ Single cell metrics for evaluating the imputation and latent space
  during training

  Parameters
  ----------
  inputs : {`SingleCellOMIC`, `numpy.ndarray`}
  extras : None
    extras object (e.g. protein) used for calculating the metric
  n_samples : `int` (default=`1`)
    number of MCMC samples for prediction
  batch_size : `int` (default=`64`)
  freq : `int` (default=`3`)
    frequency of evaluating the metric, some metrics are very computational
    intensive and could slow down the training progress significantly
  """

  def __init__(self,
               inputs: Union[SingleCellOMIC, List[SingleCellOMIC], np.
                             ndarray, List[np.ndarray], None] = None,
               extras=None,
               n_samples=1,
               batch_size=64,
               freq=3,
               name=None,
               **kwargs):
    super(SingleCellMetric, self).__init__(**kwargs)
    self.n_samples = n_samples
    self.batch_size = batch_size
    self.inputs = inputs
    self.extras = extras
    self.freq = int(freq)
    self._name = name
    # store the last epoch that the metric was calculated
    self._last_epoch = 0
    assert self.freq > 0

  @property
  def name(self):
    return self.__class__.__name__.lower() if self._name is None else self._name

  def set_model(self, model: SingleCellModel):
    assert isinstance(
        model, SingleCellModel), "This callback only support SingleCellModel"
    self.model = model
    return self

  @abstractmethod
  def call(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], extras):
    raise NotImplementedError

  def __call__(self, inputs=None, n_samples=None):
    if inputs is None:
      inputs = self.inputs
    if n_samples is None:
      n_samples = self.n_samples
    model = self.model

    if not isinstance(inputs, (tuple, list)):
      inputs = [inputs]
    inputs = [_to_sc_omics(i) for i in inputs]
    if model.corruption_rate is not None:
      corruption_text = str(model.corruption_dist) + str(model.corruption_rate)
      inputs_corrupt = [
          (data.corrupt(corruption_rate=model.corruption_rate,
                        corruption_dist=model.corruption_dist,
                        inplace=False) \
           if str(id(data)) + corruption_text not in _CORRUPTED_INPUTS else
           _CORRUPTED_INPUTS[str(id(data)) + corruption_text]) \
             if idx == 0 else data
          for idx, data in enumerate(inputs)
      ]
      _CORRUPTED_INPUTS[str(id(inputs[0])) +
                        corruption_text] = inputs_corrupt[0]
    else:
      inputs_corrupt = inputs

    outputs, latents = model.predict(inputs_corrupt,
                                     n_samples=self.n_samples,
                                     batch_size=self.batch_size,
                                     verbose=0,
                                     apply_corruption=False)
    if not isinstance(outputs, (tuple, list)):
      outputs = [outputs]
    if not isinstance(latents, (tuple, list)):
      latents = [latents]

    metrics = self.call(y_true=inputs,
                        y_pred=outputs,
                        y_crpt=inputs_corrupt,
                        latents=latents,
                        extras=self.extras)
    if metrics is None:
      metrics = {}
    elif tf.is_tensor(metrics) or \
      isinstance(metrics, np.ndarray) or \
        isinstance(metrics, Number):
      metrics = {self.name: metrics}
    assert isinstance(metrics, dict), \
      "Return metrics must be a dictionary mapping metric name to scalar value"
    metrics = {
        i: j.numpy() if tf.is_tensor(j) else j for i, j in metrics.items()
    }
    return metrics

  def on_epoch_end(self, epoch, logs=None):
    """Called at the end of an epoch.

    Subclasses should override for any actions to run. This function should only
    be called during TRAIN mode.

    Arguments:
        epoch: integer, index of epoch.
        logs: dict, metric results for this training epoch, and for the
          validation epoch if validation is performed. Validation result keys
          are prefixed with `val_`.
    """
    if epoch % self.freq == 0 and logs is not None:
      self._last_epoch = epoch
      # calculating the metric
      try:
        metrics = self()
      except Exception as e:
        print("Error:", e)
        metrics = {}
      # update the log
      for key, val in metrics.items():
        logs[key] = val
        logs[key + '_epoch'] = epoch

  def on_train_end(self, logs=None):
    if self.model.epochs != self._last_epoch:
      self._last_epoch = self.model.epochs
      # calculating the metric
      try:
        metrics = self()
      except Exception as e:
        print("Error:", e)
        metrics = {}
      # update the log
      history = self.model.history.history
      for key, val in metrics.items():
        if key in history:
          history[key].append(val)
          history[key + '_epoch'].append(self._last_epoch)


# ===========================================================================
# Losses
# ===========================================================================
class NegativeLogLikelihood(SingleCellMetric):
  """ Log likelihood metric

  Parameters
  ----------
  inputs : {`SingleCellOMIC`, `numpy.ndarray`}
  extras : None
    extras object (e.g. protein) used for calculating the metric
  n_samples : `int` (default=`1`)
    number of MCMC samples for prediction
  batch_size : `int` (default=`64`)
  freq : `int` (default=`3`)
    frequency of evaluating the metric, some metrics are very computational
    intensive and could slow down the training progress significantly

  Returns
  -------
  dict:
    'nllk%d' for each tuple of input and output
  """

  def call(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], extras):
    nllk = {}
    for idx, (t, p) in enumerate(zip(y_true, y_pred)):
      nllk['nllk%d' % idx] = -tf.reduce_mean(p.log_prob(t.X))
    return nllk


class ImputationError(SingleCellMetric):
  """ Imputation error

  Parameters
  ----------
  inputs : {`SingleCellOMIC`, `numpy.ndarray`}
  extras : None
    extras object (e.g. protein) used for calculating the metric
  n_samples : `int` (default=`1`)
    number of MCMC samples for prediction
  batch_size : `int` (default=`64`)
  freq : `int` (default=`3`)
    frequency of evaluating the metric, some metrics are very computational
    intensive and could slow down the training progress significantly

  Return
  ------
  dict :
    'imp_med'
    'imp_mean'
  """

  def call(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], extras):
    # only care about the first data input
    y_true = y_true[0]
    y_crpt = y_crpt[0]
    y_pred = y_pred[0]

    y_pred = _preprocess_output_distribution(y_pred)
    y_pred = y_pred.mean()
    if y_pred.shape.ndims == 3:
      y_pred = tf.reduce_mean(y_pred, axis=0)
    return {
        'imp_med':
            imputation_score(original=y_true.X, imputed=y_pred),
        'imp_mean':
            imputation_mean_score(original=y_true.X,
                                  corrupted=y_crpt.X,
                                  imputed=y_pred)
    }


class CorrelationScores(SingleCellMetric):
  """ (1 - correlation_coefficients) to represent the loss

  Parameters
  ----------
  inputs : {`SingleCellOMIC`, `numpy.ndarray`}
  extras : {`SingleCellOMIC`, `numpy.ndarray`}
    the protein array
  n_samples : `int` (default=`1`)
    number of MCMC samples for prediction
  batch_size : `int` (default=`64`)
  freq : `int` (default=`3`)
    frequency of evaluating the metric, some metrics are very computational
    intensive and could slow down the training progress significantly

  Returns
  -------
  dict :
    'pearson_mean': np.mean(pearson),
    'spearman_mean': np.mean(spearman),
    'pearson_med': np.median(pearson),
    'spearman_med': np.median(spearman),

  Example
  -------
  >>> CorrelationScores(extras=y_train, freq=1)

  """

  def call(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], extras):
    y_true = y_true[0]
    y_crpt = y_crpt[0]
    y_pred = y_pred[0]
    assert isinstance(extras, SingleCellOMIC), \
      "protein data must be provided as extras in form of SingleCellOMIC"
    protein = extras[y_true.indices]
    y_true.assert_matching_cells(protein)

    y_pred = _preprocess_output_distribution(y_pred)
    y_pred = y_pred.mean()
    if y_pred.shape.ndims == 3:
      y_pred = tf.reduce_mean(y_pred, axis=0)

    scores = correlation_scores(X=y_pred,
                                y=protein.X,
                                gene_name=y_true.var['geneid'],
                                protein_name=protein.var['protid'],
                                return_series=False)
    if len(scores) == 0:
      return {}
    spearman = []
    pearson = []
    for _, (s, p) in scores.items():
      spearman.append(-s)
      pearson.append(-p)
    return {
        'pearson_mean': np.mean(pearson),
        'spearman_mean': np.mean(spearman),
        'pearson_med': np.median(pearson),
        'spearman_med': np.median(spearman),
    }


class ClusteringScores(SingleCellMetric):
  """
  Parameters
  ----------
  inputs : {`SingleCellOMIC`, `numpy.ndarray`}
  extras : {`SingleCellOMIC`, `numpy.ndarray`}
    the protein array
  n_samples : `int` (default=`1`)
    number of MCMC samples for prediction
  batch_size : `int` (default=`64`)
  freq : `int` (default=`3`)
    frequency of evaluating the metric, some metrics are very computational
    intensive and could slow down the training progress significantly

  Returns
  -------
  dict :
    silhouette_score (higher is better, best is 1, worst is -1)
    adjusted_rand_score (higher is better)
    normalized_mutual_info_score (higher is better)
    unsupervised_clustering_accuracy (higher is better)

  Example
  -------
  >>> ClusteringScores(extras=y_train, freq=1)
  """

  def call(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], extras):
    y_true = y_true[0]
    y_crpt = y_crpt[0]
    y_pred = y_pred[0]
    assert isinstance(extras, SingleCellOMIC), \
      "protein data must be provided as extras in form of SingleCellOMIC"
    protein = extras[y_true.indices]
    y_true.assert_matching_cells(protein)
    labels = _to_binary(protein)

    scores = {}
    scores_avg = defaultdict(list)
    # support multiple latents also
    for idx, z in enumerate(latents):
      for key, val in clustering_scores(latent=z.mean().numpy(),
                                        labels=labels,
                                        n_labels=protein.var.shape[0]).items():
        # since all score higher is better, we want them as loss value
        val = -val
        scores['%s_%d' % (key, idx)] = val
        scores_avg[key].append(val)
    # average scores
    scores.update({i: np.mean(j) for i, j in scores_avg.items()})
    return scores



from __future__ import absolute_import, division, print_function

import base64
import os
import pickle

import numpy as np
from scipy import sparse
from six import string_types

from odin.fuel import Dataset, MmapArrayWriter
from odin.utils import as_tuple, ctext


# ===========================================================================
# Helpers
# ===========================================================================
def _check_data(X, X_col, y, y_col, rowname):
  if not sparse.issparse(X):
    assert np.min(X) >= 0, "Only support non-negative value for X"
  assert X_col.ndim == 1 and len(X_col) == X.shape[1]
  assert rowname.ndim == 1 and len(rowname) == X.shape[0]

  if y is not None and len(y.shape) > 0 and y.shape[1] != 0:
    if not sparse.issparse(y):
      assert np.min(y) >= 0, "Only support non-negative value for y"

    assert X.ndim == 2 and y.ndim == 2, "Only support matrix for `X` and `y`"
    assert X.shape[0] == y.shape[0], \
    "Number of sample mismatch `X=%s` and `y=%s`" % (X.shape, y.shape)
    assert y_col.ndim == 1 and len(y_col) == y.shape[1]


def read_gzip_csv(path):
  import gzip
  with gzip.open(path, 'rb') as file_obj:
    data = []
    for line in file_obj:
      line = str(line, 'utf-8').strip()
      line = line.split(',')
      data.append(line)
    data = np.array(data)
    return data


# ===========================================================================
# Some normalization for meta-data
# ===========================================================================
_protein_name = {
    "PD-L1;CD274": "CD274",
    "PECAM;CD31": "CD31",
    "CD26;Adenosine": "CD26",
    "CD366;tim3": "CD366",
    "MHCII;HLA-DR": "MHCII",
    "IL7Ralpha;CD127": "CD127",
    "PD-1": "PD-1",  # CD279
    "PD1": "PD1",
    "B220;CD45R": "CD45R",
    "Ox40;CD134": "CD134",
    "CD8a": "CD8",
    "CD8A": "CD8",
    "CD4 T cells": "CD4",
    "CD8 T cells": "CD8",
}


def standardize_protein_name(name):
  """ standardize """
  assert isinstance(name, string_types), "Protein name must be string types"
  name = name.replace('-TotalSeqB', '')
  name = name.strip()
  # regular expression could be used but it could brutally remove
  # a lot of things without our notice
  if name in _protein_name:
    name = _protein_name[name]
  return name


# ===========================================================================
# Gene identifier processing
# ===========================================================================
def get_gene_id2name():
  """ Return the mapping from gene identifier to gene symbol (i.e. name)
  for PBMC 8k data
  """
  from odin.utils import get_file
  from sisua.data.path import DOWNLOAD_DIR
  url = base64.decodebytes(
      b'aHR0cHM6Ly9haS1kYXRhc2V0cy5zMy5hbWF6b25hd3MuY29tL2dlbmVfaWQybmFtZS5wa2w=\n'
  )
  url = str(url, 'utf-8')
  get_file('gene_id2name.pkl', url, DOWNLOAD_DIR)
  with open(os.path.join(DOWNLOAD_DIR, 'gene_id2name.pkl'), 'rb') as f:
    return pickle.load(f)


# ===========================================================================
# Utilities
# ===========================================================================
def remove_allzeros_columns(matrix, colname, print_log=True):
  """ Remove all zero columns from both the matrix and column name vector

  Return
  ------
  matrix : [n_samples, n_genes]
  column_name : [n_genes]
  """
  assert matrix.ndim == 2
  orig_shape = matrix.shape
  # at least > 1 for train, test splitting
  nonzero_col = np.sum(matrix, axis=0) > 1
  matrix = matrix[:, nonzero_col]
  colname = colname[nonzero_col]
  if print_log:
    print("Filtering %d all-zero columns from data: %s -> %s ..." %
          (len(nonzero_col) - np.sum(nonzero_col), str(orig_shape),
           str(matrix.shape)))
  return matrix, colname


def validating_dataset(path):
  if isinstance(path, Dataset):
    ds = path
  elif isinstance(path, string_types):
    ds = Dataset(path, read_only=True)

  assert 'X' in ds, \
  '`X` (n_samples, n_genes) must be stored at path: %s' % ds.path
  assert 'X_col' in ds, \
  '`X_col` (n_genes,) must be stored at path: %s' % ds.path
  assert 'X_row' in ds, \
  '`X_row` (n_samples,) must be stored at path: %s' % ds.path

  if 'y' in ds:
    assert 'y' in ds, \
    '`y` (n_samples, n_protein) must be stored at path: %s' % ds.path
    assert 'y_col' in ds, \
    '`y_col` (n_protein,) must be stored at path: %s' % ds.path
    y, y_col = ds['y'], ds['y_col']
  else:
    y, y_col = None, None

  X, X_col, rowname = \
  ds['X'], ds['X_col'],  ds['X_row']
  _check_data(X, X_col, y, y_col, rowname)


def save_to_dataset(path,
                    X,
                    X_col=None,
                    y=None,
                    y_col=None,
                    rowname=None,
                    print_log=True):
  """
  path : output folder path
  X : (n_samples, n_genes) gene expression matrix
  X_col : (n_genes,) name of each gene
  y : (n_samples, n_proteins) protein marker level matrix
  y_col : (n_proteins) name of each protein
  rowname : (n_samples,) name of cells (i.e. the sample)
  print_log : bool (default: True)
  """
  _check_data(X, X_col, y, y_col, rowname)
  assert os.path.isdir(path), "'%s' must be path to a folder" % path
  # save data
  if print_log:
    print("Saving data to %s ..." % ctext(path, 'cyan'))
  # saving sparse matrix
  if sparse.issparse(X):
    with open(os.path.join(path, 'X'), 'wb') as f:
      pickle.dump(X, f)
  else:
    with MmapArrayWriter(path=os.path.join(path, 'X'),
                         dtype='float32',
                         shape=(0, X.shape[1]),
                         remove_exist=True) as out:
      out.write(X)
  # save the meta info (X features)
  if X_col is not None:
    with open(os.path.join(path, 'X_col'), 'wb') as f:
      pickle.dump(X_col, f)
  # saving the label data (can be continous or discrete or binary)
  if y is not None and len(y.shape) > 0 and y.shape[1] != 0:
    if sparse.issparse(y):
      with open(os.path.join(path, 'y'), 'wb') as f:
        pickle.dump(y, f)
    else:
      with MmapArrayWriter(path=os.path.join(path, 'y'),
                           dtype='float32',
                           shape=(0, y.shape[1]),
                           remove_exist=True) as out:
        out.write(y)
    with open(os.path.join(path, 'y_col'), 'wb') as f:
      pickle.dump(y_col, f)
  # row name for both X and y
  if rowname is not None:
    with open(os.path.join(path, 'X_row'), 'wb') as f:
      pickle.dump(rowname, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import pip
from setuptools import setup, find_packages

_SISUA_VERSION = '0.4.3'

if not (sys.version_info.major == 3 and sys.version_info.minor == 6):
  raise RuntimeError("Sorry, we only support Python=3.6!")

if float(pip.__version__.split('.')[0]) < 19.0:
  raise RuntimeError(
      "'sisua' package require pip version >= 19.0, your pip version is %s, "
      "run `pip install pip --upgrade` to upgrade!" % str(pip.__version__))

# ===========================================================================
# Main
# ===========================================================================
with open('README.rst') as readme_file:
  readme = readme_file.read()

author = 'University of Eastern Finland'

requirements = [
    "odin-ai==1.1.1",
    "seaborn>=0.9",
    "pandas",
    'scanpy==1.4.4',
]

setup(
    author=author,
    author_email='trung@imito.ai',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
    ],
    description="SemI-SUpervised generative Autoencoder for single cell data",
    long_description=readme,
    long_description_content_type='text/x-rst',
    scripts=['bin/sisua-train',
             'bin/sisua-analyze'],
    setup_requires=['pip>=19.0'],
    install_requires=requirements,
    license="MIT license",
    include_package_data=True,
    keywords='sisua',
    name='sisua',
    packages=find_packages(),
    test_suite='tests',
    url='https://github.com/trungnt13/sisua',
    version=_SISUA_VERSION,
    zip_safe=False,
)



from matplotlib import pyplot as plt
import seaborn
import pandas as pd

import numpy as np
from scipy.stats import kde, entropy

from odin import backend as K, visual
from odin.visual import plot_series_statistics
from odin.utils import catch_warnings_ignore
from odin.stats import describe, sparsity_percentage

# ===========================================================================
# Some helper
# ===========================================================================
def _mean(W_or_V):
  """ The reconstruction from VAE is returned by:
  (mean, total_stddev, explained_stddev)
  This method will make sure only mean value is return
  """
  W_or_V = (W_or_V[0]
          if isinstance(W_or_V, (tuple, list)) or W_or_V.ndim == 3 else
          W_or_V)
  assert W_or_V.ndim == 2
  return W_or_V


# ===========================================================================
# Series analysis
# ===========================================================================
def _show_zero_inflated_pi(p, ax, handles, indices):
  ax = ax.twinx()
  _, = ax.plot(p[indices],
               marker='o', linestyle='None',
               markersize=0.8, color='blue',
               alpha=0.3, label="$p_{dropout}$")
  ax.set_ylabel("Zero-inflation rate", fontsize=8)
  ax.tick_params(axis='both', labelsize=8)
  handles.append(_)

def plot_countsum_series(original, imputed, p=None,
                         reduce_axis=0, title=None, ax=None):
  """
  x: [n_samples, n_genes]
    original count
  w: tuple (expected, stdev_total, stdev_explained) [n_samples, n_genes]
    the prediction
  p: [n_samples, n_genes]
    dropout probability
  """
  if ax is None:
    ax = visual.to_axis(ax)
  reduce_axis = int(reduce_axis)

  if isinstance(imputed, (tuple, list)): # no statistics provided
    assert len(imputed) == 3
    expected, stdev_total, stdev_explained = imputed
  elif imputed.ndim == 3:
    assert imputed.shape[0] == 3
    expected, stdev_total, stdev_explained = imputed[0], imputed[1], imputed[2]
  else:
    raise ValueError()

  count_sum_observed = np.log1p(np.sum(original, axis=reduce_axis))
  count_sum_expected = np.log1p(np.sum(expected, axis=reduce_axis))
  count_sum_stdev_total = np.log1p(np.sum(stdev_total, axis=reduce_axis))
  count_sum_stdev_explained = np.log1p(np.sum(stdev_explained, axis=reduce_axis))
  if p is not None:
    p_sum = np.mean(p, axis=reduce_axis)

  ax, handles, indices = plot_series_statistics(
      count_sum_observed, count_sum_expected,
      explained_stdev=count_sum_stdev_explained,
      total_stdev=count_sum_stdev_total,
      fontsize=8, ax=ax, legend_enable=False,
      title=title,
      despine=True if p is None else False,
      return_handles=True, return_indices=True,
      xscale='linear', yscale='linear', sort_by='expected')
  if p is not None:
    _show_zero_inflated_pi(p_sum, ax, handles, indices)
  ax.legend(handles=handles, loc='best', markerscale=4, fontsize=8)

# ===========================================================================
# Comparing Gene expression
# ===========================================================================
def plot_countsum_comparison(original, reconstructed, imputed, title,
                             comparing_axis=0, ax=None):
  """
  original : [n_samples, n_genes]
  reconstructed : [n_samples, n_genes]
  imputed : [n_samples, n_genes]
  """
  from matplotlib import pyplot as plt
  ax = visual.to_axis(ax)

  original = original.sum(axis=comparing_axis)
  reconstructed = _mean(reconstructed).sum(axis=comparing_axis)
  imputed = _mean(imputed).sum(axis=comparing_axis)
  assert original.shape == reconstructed.shape == imputed.shape

  sorted_indices = np.argsort(original)

  original = np.log1p(original[sorted_indices])
  reconstructed = np.log1p(reconstructed[sorted_indices])
  imputed = np.log1p(imputed[sorted_indices])

  # ====== plotting the figures ====== #
  colors = seaborn.color_palette(palette='Set2', n_colors=3)

  ax.scatter(original, imputed, c=colors[1], s=3, alpha=0.3)
  ax.scatter(original, reconstructed, c=colors[2], s=3, alpha=0.3)
  # ====== plotting the median line ====== #
  xmin, xmax = ax.get_xlim()
  ymin, ymax = ax.get_ylim()
  max_val = max(xmax, ymax)

  ax.axhline(xmin=0, xmax=max_val, y=np.median(original),
             color=colors[0], linestyle='--', linewidth=1.5,
             label="Corrupted Median")
  ax.axhline(xmin=0, xmax=max_val, y=np.median(imputed),
             color=colors[1], linestyle='--', linewidth=1.5,
             label="Imputed Median")
  ax.axhline(xmin=0, xmax=max_val, y=np.median(reconstructed),
             color=colors[2], linestyle='--', linewidth=1.5,
             label="Reconstructed Median")
  # ====== adjust the aspect ====== #
  visual.plot_aspect(aspect='equal', adjustable='box', ax=ax)
  ax.set_xlim((0, max_val))
  ax.set_ylim((0, max_val))

  ax.plot((0, max_val), (0, max_val),
          color='black', linestyle=':', linewidth=1)
  plt.legend(fontsize=8)
  ax.set_xlabel("Log-Count of the corrupted data")
  ax.set_ylabel("Log-Count of the reconstructed and imputed data")
  ax.set_title(title)

  return ax



from __future__ import print_function, absolute_import, division

import os
import numpy as np

from scvi.dataset import GeneExpressionDataset
from sisua.data import get_dataset

class FacsDataset(GeneExpressionDataset):

  def __init__(self, n_proteins=7):
    assert n_proteins in (2, 5, 7), "Only support: 2, 5 or 7 protein FACS dataset"

    self.n_proteins = int(n_proteins)
    expression_data = self.download_and_preprocess()
    super().__init__(
        *GeneExpressionDataset.get_attributes_from_matrix(expression_data)
    )

  def preprocess(self):
    ds, gene_ds, prot_ds = get_dataset(
        dataset_name="facs_%d" % self.n_proteins, override=False)
    expression_data = gene_ds.X
    gene_symbols = gene_ds.X_col
    self.gene_symbols = gene_symbols
    self.cell_names = gene_ds.X_row

    self.adt_expression = prot_ds.X
    self.protein_markers = prot_ds.X_col

    assert np.all(gene_ds.X_row == prot_ds.X_row)

    return expression_data


class PbmcCiteseqDataset(GeneExpressionDataset):

  def __init__(self):
    expression_data = self.download_and_preprocess()
    super().__init__(
        *GeneExpressionDataset.get_attributes_from_matrix(expression_data)
    )

  def preprocess(self):
    ds, gene_ds, prot_ds = get_dataset(dataset_name="pbmc_citeseq",
                                       override=False)
    expression_data = gene_ds.X
    gene_symbols = gene_ds.X_col
    self.gene_symbols = gene_symbols
    self.cell_names = gene_ds.X_row

    self.adt_expression = prot_ds.X
    self.protein_markers = prot_ds.X_col

    assert np.all(gene_ds.X_row == prot_ds.X_row)

    return expression_data



import base64
import os
import pickle
import shutil
from io import BytesIO

import numpy as np

from odin.fuel import Dataset
from odin.utils import batching, ctext, get_file, select_path
from odin.utils.crypto import decrypt_aes, md5_checksum
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

# ===========================================================================
# Constants
# ===========================================================================
# Protein
_URL_LYMPHOID = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL3BibWM4a19seS5ucHo=\n'
_URL_MYELOID = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL3BibWM4a19teS5ucHo=\n'
_URL_PBMC8k = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL3BibWM4a19mdWxsLm5weg==\n'


# ===========================================================================
# Main
# ===========================================================================
def read_PBMC8k(subset, override=False, verbose=False, filtered_genes=False):
  subset = str(subset).strip().lower()
  if subset not in ('ly', 'my', 'full'):
    raise ValueError(
        "subset can only be 'ly'-lymphoid and 'my'-myeloid or 'full'")

  download_path = os.path.join(DOWNLOAD_DIR, "PBMC8k_%s_original" % subset)
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  preprocessed_path = os.path.join(
      PREPROCESSED_BASE_DIR,
      'PBMC8k_%s_preprocessed' % (subset + ('' if filtered_genes else 'full')))

  if override and os.path.exists(preprocessed_path):
    shutil.rmtree(preprocessed_path)
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)

  # ******************** preprocessed ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    # ====== pbmc 8k ====== #
    if subset == 'full':
      ly = read_PBMC8k('ly', override=override, filtered_genes=filtered_genes)
      my = read_PBMC8k('my', override=override, filtered_genes=filtered_genes)

      url = str(base64.decodebytes(_URL_PBMC8k), 'utf-8')
      base_name = os.path.basename(url)
      get_file(fname=base_name,
               origin=url,
               outdir=download_path,
               verbose=verbose)

      data = np.load(os.path.join(download_path, base_name))
      X = data['X']
      X_row = data['X_row']
      X_col = data['X_col'].tolist()
      y = data['y']
      y_col = data['y_col'].tolist()

      all_genes = set(ly['X_col'].tolist() + my['X_col'].tolist())
      all_genes = sorted([X_col.index(i) for i in all_genes])

      all_proteins = set(ly['y_col'].tolist() + my['y_col'].tolist())
      all_proteins = sorted([y_col.index(i) for i in all_proteins])

      X = X[:, all_genes]
      y = y[:, all_proteins]
      X_col = np.array(X_col)[all_genes]
      y_col = np.array(y_col)[all_proteins]
      cell_types = np.array(['ly' if i in ly['X_row'] else 'my' for i in X_row])
    # ====== pbmc ly and my ====== #
    else:
      url = str(
          base64.decodebytes(_URL_LYMPHOID if subset == 'ly' else _URL_MYELOID),
          'utf-8')
      base_name = os.path.basename(url)
      get_file(fname=base_name,
               origin=url,
               outdir=download_path,
               verbose=verbose)
      # ====== extract the data ====== #
      data = np.load(os.path.join(download_path, base_name))
      X_row = data['X_row']
      y = data['y']
      y_col = data['y_col']
      if filtered_genes:
        X = data['X_filt']
        X_col = data['X_filt_col']
      else:
        X = data['X_full']
        X_col = data['X_full_col']
      cell_types = None

    # ====== save everything ====== #
    X, X_col = remove_allzeros_columns(matrix=X,
                                       colname=X_col,
                                       print_log=verbose)
    assert X.shape == (len(X_row), len(X_col))
    assert len(X) == len(y)
    assert y.shape[1] == len(y_col)

    if cell_types is not None:
      with open(os.path.join(preprocessed_path, 'cell_types'), 'wb') as f:
        pickle.dump(cell_types, f)
    save_to_dataset(preprocessed_path,
                    X,
                    X_col,
                    y,
                    y_col,
                    rowname=X_row,
                    print_log=verbose)

  # ******************** read preprocessed data ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



from __future__ import absolute_import, division, print_function

from collections import OrderedDict, defaultdict

import numpy as np
import pandas as pd
import seaborn
from matplotlib import pyplot as plt
from scipy.stats import kde
from six import string_types
from sklearn.neighbors import KernelDensity

from odin import backend as K
from odin import visual
from odin.utils import as_tuple, cache_memory, catch_warnings_ignore
from odin.visual import plot_figure, to_axis
from sisua.data import get_dataset
from sisua.data.const import MARKER_GENES
from sisua.data.utils import standardize_protein_name


# ===========================================================================
# Helpers
# ===========================================================================
def get_imputed_indices(x_org, x_imp):
  """ Return the indices of the cells which are imputed"""
  ids = []
  for i, (xo, xi) in enumerate(zip(x_org, x_imp)):
    if np.sum(xo) != np.sum(xi):
      ids.append(i)
  return np.array(ids)


def correlation_scores(X, y, gene_name, protein_name, return_series=False):
  """ Spearman and Pearson correlation scores

  return_series : bool
      if True, return the gene and protein series, instead of
      calculating spearman, or pearson correlation scores

  Return
  ------
  `correlation_scores` or `correlation_series`

  of which,

  correlation_scores : dict(protein_name/gene_name=(spearman, pearson))
  correlation_series : dict(protein_name/gene_name=(gene_series, prot_series))

  """
  from scipy.stats import pearsonr, spearmanr

  gene_name = np.asarray(gene_name).ravel().tolist()
  protein_name = np.asarray(protein_name).ravel().tolist()

  assert X.shape[1] == len(gene_name), "Number of genes mismatch"
  assert y.shape[1] == len(protein_name), "Number of protein mismatch"
  assert X.shape[0] == y.shape[0], "Number of sample mismatch"

  # NOTE: the process of filtering gene using variable genes might
  # remove marker gene of the protein

  # pre-processing protein name
  protein_name = [standardize_protein_name(i) for i in protein_name]
  # mapping from protein index to gene index
  prot2gene = {}
  for prot, gene in MARKER_GENES.items():
    if prot in protein_name:
      index = [i for i, name in enumerate(gene_name) if gene == name]
      if len(index) == 0:  # still not found anything
        index = [
            i for i, name in enumerate(gene_name) if gene == name.split('_')[-1]
        ]
      if len(index) == 1:  # found
        prot2gene[protein_name.index(prot)] = index[0]
      elif len(index) >= 2:
        raise RuntimeError("Found multiple gene index with the same name")

  # get the series and calculating scores
  scores = {}
  series = {}
  if len(prot2gene) > 0:
    for prot, gene in prot2gene.items():
      x_ = X[:, gene]
      y_ = y[:, prot]
      name = protein_name[prot] + '/' + gene_name[gene]
      series[name] = (x_, y_)
      if not return_series:
        scores[name] = (spearmanr(x_, y_).correlation, pearsonr(x_, y_)[0])

  if return_series:
    series = OrderedDict(
        sorted(series.items(), key=lambda x: x[0].split('/')[0]))
    return series
  else:
    scores = OrderedDict(
        sorted(scores.items(), key=lambda x: x[0].split('/')[0]))
    return scores


# ===========================================================================
# Metrics
# ===========================================================================
def imputation_score(original, imputed):
  """ Median of medians for all distances """
  assert original.shape == imputed.shape
  nonzeros = np.nonzero(original)
  d = np.abs(original - imputed)  # [nonzeros]
  return float(np.median(d))


def imputation_mean_score(original, corrupted, imputed):
  """ Mean of medians for each cell imputation score """
  assert original.shape == corrupted.shape == imputed.shape
  imputation_cells = []
  for cell_org, cell_crt, cell_imp in zip(original, corrupted, imputed):
    if np.sum(cell_org) != np.sum(cell_crt):
      imputation_cells.append(np.median(np.abs(cell_org - cell_imp)))
  return np.mean(imputation_cells) if len(imputation_cells) > 0 else 0


def imputation_std_score(original, corrupted, imputed):
  # Standard deviation of medians for each cell imputation score
  assert original.shape == corrupted.shape == imputed.shape
  imputation_cells = []
  for cell_org, cell_crt, cell_imp in zip(original, corrupted, imputed):
    if np.sum(cell_org) != np.sum(cell_crt):
      imputation_cells.append(np.median(np.abs(cell_org - cell_imp)))
  return np.std(imputation_cells) if len(imputation_cells) > 0 else 0


# ===========================================================================
# Imputation analysis
# ===========================================================================
def plot_imputation_series(original, imputed, title="Imputation"):
  original = K.log_norm(original, axis=0)
  imputed = K.log_norm(imputed, axis=0)
  max_val = max(np.max(original), np.max(imputed))

  with catch_warnings_ignore(FutureWarning):
    grid = seaborn.pairplot(data=pd.DataFrame({
        'Original Value': original,
        'Imputed Value': imputed
    }),
                            height=4,
                            aspect=1,
                            kind='reg',
                            diag_kws={'bins': 180},
                            plot_kws={
                                'scatter_kws': dict(s=2, alpha=0.6),
                                'line_kws': dict(color='red', alpha=0.8),
                                'color': 'g'
                            })
    ids = np.linspace(0, max_val)
    grid.axes[0, 1].set_xlim((0, max_val))
    grid.axes[0, 1].set_ylim((0, max_val))
    grid.axes[0, 1].plot(ids, ids, linestyle='--', linewidth=1, color='black')

    grid.axes[1, 0].set_xlim((0, max_val))
    grid.axes[1, 0].set_ylim((0, max_val))
    grid.axes[1, 0].plot(ids, ids, linestyle='--', linewidth=1, color='black')


def plot_imputation(original,
                    imputed,
                    corrupted=None,
                    kde_kernel='gaussian',
                    ax=None,
                    title="Imputation"):
  """ Original code: scVI
      Modified by: SISUA

  kde_kernel : string (default: 'linear')
    'gaussian', 'tophat', 'epanechnikov', 'exponential', 'linear',
    'cosine'

  """
  y = imputed
  x = original
  assert imputed.shape == original.shape
  if corrupted is not None:
    assert original.shape == corrupted.shape

  # this modification focus only on missing data points
  if corrupted is not None:
    mask = np.where(original != corrupted, True, False)
    x = x[mask]
    y = y[mask]

  ymax = 25  # increasing ymax for more data points
  #
  mask = x < ymax
  x = x[mask]
  y = y[mask]
  #
  mask = y < ymax
  x = x[mask]
  y = y[mask]
  #
  l_minimum = np.minimum(x.shape[0], y.shape[0])
  x = x[:l_minimum]
  y = y[:l_minimum]

  data = np.vstack([x, y])

  axes = visual.to_axis(ax)
  axes.set_xlim([0, ymax])
  axes.set_ylim([0, ymax])

  nbins = 80

  # Evaluate a gaussian kde on a regular grid of nbins x nbins over data extents
  xi, yi = np.mgrid[0:ymax:nbins * 1j, 0:ymax:nbins * 1j]
  # ====== scipy ====== #
  k_ = kde.gaussian_kde(data)
  zi = k_(np.vstack([xi.flatten(), yi.flatten()]))
  # ====== sklearn ====== #
  # k_ = KernelDensity(kernel=str(kde_kernel))
  # k_.fit(data.T)
  # zi = k_.score_samples(
  #     np.vstack([xi.flatten(), yi.flatten()]).T).T

  plt.title(title, fontsize=12)
  plt.ylabel("Imputed counts")
  plt.xlabel('Original counts')

  plt.pcolormesh(yi, xi, zi.reshape(xi.shape), cmap="Reds")

  a, _, _, _ = np.linalg.lstsq(y[:, np.newaxis], x, rcond=-1)
  linspace = np.linspace(0, ymax)
  plt.plot(linspace, a * linspace, color='black')

  plt.plot(linspace, linspace, color='black', linestyle=":")
